/**
 * Example of how to call the TCB DistributedFailureDetection service.
 * 
 * This client demo is responsible for starting distributed timing failure detections
 * (and the associated measurements) that are ended by server demo. 
 * Only server demo presents the associated measurements results.
 * 
 * A thread is launched to monitor the results of distributed timing 
 * failure detections.
 *
 * Another thread is launched to react to timing failures.
 * 
 * To call this service, you need first to create a DLL with
 * two functions:
 * 
 * 		1) The function to timely execute if a failure occurs
 * 		2) A function that returns the Worst Case Execution Time
 * 		   (milliseconds) of the function specified in 1). 
 * 
 * There is already a dll example created that works with this example.
 * The source code (Embedded Visual C++ 3.0) is located in directory
 * "tcb_services_examples/dll_example_for_tcb_handler". You should compile it
 * and deploy in PocketPC prior to running this example.
 * 
 * @author Paulo Sousa
 */

#include "stdafx.h"
//include sockets header file
#include <winsock.h>
//include tcb services library header file
#include "..\..\..\..\tcb_services_library\tcb_services_library.h"


//TODO: replace this values according to your situation
#define MY_IP "192.168.0.1"
#define DEST_IP "224.0.0.10"
#define DEST_PORT 40000
//target TCB event channel
#define EVENT_CHANNEL_ID 10


/**
 * Task that reacts to a timing failure
 */
DWORD WINAPI fail_task (LPVOID param) {
	//create event to receive failure notification
	HANDLE failure_event = CreateEvent(NULL, TRUE, FALSE, L"failure_event");
	if (failure_event==NULL) {
		printf("Error creating failure event\n");
	}
	int count=0;
	while(1) {
		WaitForSingleObject(failure_event, INFINITE);
		ResetEvent(failure_event);
		printf("***failure handler count=%d***\n", count++);
	}
	CloseHandle(failure_event);
	return 0;
}	

/**
 * Task that prints distributed FD info
 */
DWORD WINAPI wait_info_task (LPVOID param) {
	tcb_distributed_fd_info_t info;
	
	while(1) {
	 
	 //wait for distributed FD info
	 info = tcb_wait_info();
	 if (info.from==-1) {
		printf("Error: TCB session is not opened\n");
		return 0;
	 }

	 printf("--------------------\n");
	 		
	 //print info sender and sequence number
	 struct in_addr in;
	 in.S_un.S_addr = info.from;
	 printf("W_INFO:Info from %s, seq=%d\n", inet_ntoa(in), info.seq);
	 //print fail info for each machine (per tcb id)
	 for (int i=0; i<MAX_NUMBER_OF_MACHINES; i++) {
		if (info.failInfo[i].empty)
			continue;
		
		//if there exists no info
		if (info.failInfo[i].fail==2) {
			struct in_addr in;
			in.S_un.S_addr = info.failInfo[i].id;
			printf("W_INFO:Id %s: NO INFO\n",inet_ntoa(in));
		}
		//if there exists info
		else {
			struct in_addr in;
			in.S_un.S_addr = info.failInfo[i].id;
			printf("W_INFO:Id %s: ",inet_ntoa(in));
			//verify if a timing failure occured
			if (info.failInfo[i].fail)
				printf("failed");
			else printf("timely");
			//if lateness is available
			if (info.failInfo[i].latenessAvail)
				printf(", lat.degree=%d ms\n", info.failInfo[i].latenessDegree);
			else printf(", no lateness available\n");
		}
	 }
	}
	
	return 0;
}




int main(int argc, char* argv[]) {

	//open TCB session
	printf("Opening TCB session\n");
	int error = open_tcb_session();
	if (error!=0) {
		printf("Error opening session=%d\n",error);
		getchar();
		return -1;
	}

	
	//launch wait_info thread
	HANDLE task_id =
		CreateThread(NULL,0,wait_info_task,NULL,NULL,NULL);
	if (task_id==NULL){
		printf("FATAL ERROR: Couldn't start wait_info task\n");
		getchar();
		return -1;
	}
	

	//launch fail_task thread
	HANDLE task_id2 =
		CreateThread(NULL,0,fail_task,NULL,NULL,NULL);
	if (task_id2==NULL){
		printf("FATAL ERROR: Couldn't start fail task\n");
		getchar();
		return -1;
	}
	//Give a higher priority to this thread
	CeSetThreadPriority(task_id2, 5);

	//create a socket
	printf("Creating socket\n");
	SOCKET s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s==INVALID_SOCKET) {
		printf("Error creating socket: %d\n", error);
		getchar();
		return -1;
	}

	//bind socket
	printf("Binding socket\n");
	struct sockaddr_in sa;
	sa.sin_family=AF_INET;
	sa.sin_addr.s_addr= inet_addr(MY_IP);
	sa.sin_port=htons(0);
	error= bind(s, (struct sockaddr*)&sa, sizeof(sa));
	if (error!=0) {
		printf("Error binding socket: %d\n", error);
		getchar();
		return -1;
	}

	//timing failure detection specification
	tcb_dist_action_id_t action_id;
	//time to failure (ms)
	int spec = 100;
	//maximum time allowed for the execution of failure handler (ms)
	int deadline = 0;
	char dllName[100];
	char funcName[100];
	char wcetName[100];
	//DLL where resides the failure handler and the wcet function
	strcpy(dllName, "tcb_handler_example.dll");
	//failure handler name
	strcpy(funcName, "distFD_example");
	//wcet name
	strcpy(wcetName, "distFD_example_wcet");
	int nbytes=0;

	//start 1000 services, one per each 5 seconds
	printf("Starting dist. FDs\n");
	for (int i=0; i<1000; i++) {
	 int start_ev = get_tcb_timestamp();
	 error=start_tcb_dist_fd(start_ev, EVENT_CHANNEL_ID, spec, deadline, dllName, 
							funcName, wcetName, &action_id);
	 if (error!=0) {
		printf("Error starting dist fd:");
		switch (error) {
			case -1: printf("no TCB session opened!\n");
					 break;
			case -2: printf("DLL not found!\n");
					 break;
			case -3: printf("Function not found!\n");
					 break;
			case -4: printf("WCET Function not found!\n");
					 break;
			case -5: printf("Too many dist fds running!\n");
					 break;
			case -6: printf("Timing failure already occured!\n");
					 break;
			case -7: printf("Not possible to execute dist fd!\n");
					 break;
			case -8: printf("No TCB subscribed to event channel specified!\n");
					 break;
			default: printf("unknown error\n");
		}
		Sleep(5000);
		continue;
	 }
	
	 //send a msg and serialize the identifier of the dist fd in it
	 char buf[50+sizeof(action_id)];
	 //action_id goes in first sizeof(action_id) bytes
	 memcpy(buf, (char*)&action_id, sizeof(action_id));
	 strcpy(buf+sizeof(action_id), "Hi server (FD)!");
	 //target socket
	 struct sockaddr_in to;
	 to.sin_family = AF_INET;
	 to.sin_addr.s_addr = inet_addr(DEST_IP);
	 to.sin_port = htons(DEST_PORT);
		
	 nbytes = sendto(s, buf, 50+sizeof(action_id), NULL, (struct sockaddr*)&to, 
						    sizeof(to));
		
	 if (nbytes<0) {
		printf("Error sending msg: %d\n", nbytes);
	 }
	 else printf("%d bytes sent, spec=%d\n", nbytes, spec);
	 
	 //sleep 5 seconds between each service 
	 Sleep(5000);	
	}   

	//close socket
	printf("Closing socket\n");
	error = closesocket(s);
	if (error!=0) {
		printf("Error closing socket=%d\n",error);
		getchar();
		return -1;
	}

	//close TCB session
	printf("Closing TCB session\n");
	error=close_tcb_session();
	if (error!=0) {
		printf("Error closing session=%d\n",error);
		getchar();
		return -1;
	}

	//wait for user to press a key
	printf("Press a key to close window...\n");
	getchar();
	return 0;
}
