/**
 * Example of how to call the TCB DistributedRemoteFailureDetection service.
 * 
 * This client demo is responsible for starting distributed remote failure detections
 * (and the associated measurements) that are ended by server demo. 
 * Only server demo presents the associated measurements results.
 * 
 * A thread is launched to monitor the results of distributed timing 
 * failure detections.
 *
 * @author Paulo Sousa
 */

#include "stdafx.h"
//include sockets header file
#include <winsock.h>
//include tcb services library header file
#include "..\..\..\..\tcb_services_library\tcb_services_library.h"


//TODO: replace this values according to your situation
#define MY_IP "192.168.0.1"
#define DEST_IP "224.0.0.5"
#define DEST_PORT 40000
//target event channel
#define EVENT_CHANNEL_ID 5


/**
 * Task that prints distributed FD info
 */
DWORD WINAPI wait_info_task (LPVOID param) {
	tcb_distributed_fd_info_t info;
	
	while(1) {
	 
	 //wait for distributed FD info
	 info = tcb_wait_info();
	 if (info.from==-1) {
		printf("Error: TCB session is not opened\n");
		return 0;
	 }
	 printf("--------------------\n");
	 		
	 //print info sender and sequence number
	 struct in_addr in;
	 in.S_un.S_addr = info.from;
	 printf("W_INFO:Info from %s, seq=%d\n", inet_ntoa(in), info.seq);
	 //print fail info for each machine (per tcb id)
	 for (int i=0; i<MAX_NUMBER_OF_MACHINES; i++) {
		if (info.failInfo[i].empty)
			continue;
		
		//if there exists no info
		if (info.failInfo[i].fail==2) {
			struct in_addr in;
			in.S_un.S_addr = info.failInfo[i].id;
			printf("W_INFO:Id %s: NO INFO\n",inet_ntoa(in));
		}
		//if there exists info
		else {
			struct in_addr in;
			in.S_un.S_addr = info.failInfo[i].id;
			printf("W_INFO:Id %s: ",inet_ntoa(in));
			//verify if a timing failure occured
			if (info.failInfo[i].fail)
				printf("failed");
			else printf("timely");
			//if lateness is available
			if (info.failInfo[i].latenessAvail)
				printf(", lat.degree=%d ms\n", info.failInfo[i].latenessDegree);
			else printf(", no lateness available\n");
		}
	 }
	}
	
	return 0;
}


int main(int argc, char* argv[]) {

	//open_session
	printf("Opening TCB session\n");
	int error = open_tcb_session();
	if (error!=0) {
		printf("Error opening session=%d\n",error);
		getchar();
		return -1;
	}

	//launch wait_info thread
	HANDLE task_id =
		CreateThread(NULL,0,wait_info_task,NULL,NULL,NULL);
	if (task_id==NULL){
		printf("FATAL ERROR: Couldn't start wait_info task\n");
		getchar();
		return -1;
	}


	//create a socket
	printf("Creating socket\n");
	SOCKET s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s==INVALID_SOCKET) {
		printf("Error creating socket: %d\n", error);
		getchar();
		return -1;
	}

	//bind socket
	printf("Binding socket\n");
	struct sockaddr_in sa;
	sa.sin_family=AF_INET;
	sa.sin_addr.s_addr= inet_addr(MY_IP);
	sa.sin_port=htons(0);
	error= bind(s, (struct sockaddr*)&sa, sizeof(sa));
	if (error!=0) {
		printf("Error binding socket: %d\n", error);
		getchar();
		return -1;
	}

	tcb_dist_action_id_t action_id;
	//time to failure (ms)
	int spec = 200;
				
	//start 1000 services, one per each 5 seconds
	printf("Starting remote dist. FDs\n");
	for (int i=0; i<1000; i++) {
		error=start_tcb_dist_remote_fd(get_tcb_timestamp(), EVENT_CHANNEL_ID, spec, 
								   &action_id);
		if (error!=0) {
			printf("Error starting dist remote fd:");
			switch (error) {
				case -1: printf("no TCB session opened!\n");
					 break;
				case -5: printf("Too many dist fds running!\n");
					 break;
				case -6: printf("Timing failure already occured!\n");
					 break;
				case -7: printf("Not possible to execute dist fd!\n");
					 break;
				case -8: printf("No TCB is subscribed to event channel specified!\n");
					 break;
				default: printf("unknown: %d\n", error);
			}
			Sleep(5000);
			continue;
		}

		//send a msg and serialize the identifier of the dist remote fd in it
		char buf[50+sizeof(action_id)];
		//action_id goes in first sizeof(action_id) bytes
		memcpy(buf, (char*)&action_id, sizeof(action_id));
		strcpy(buf+sizeof(action_id), "Hi server (remote FD)!");
		struct sockaddr_in to;
		to.sin_family = AF_INET;
		to.sin_addr.s_addr = inet_addr(DEST_IP);
		to.sin_port = htons(DEST_PORT);
		
		int nbytes = sendto(s, buf, 50+sizeof(action_id), NULL, (struct sockaddr*)&to, 
						    sizeof(to));
		
		if (nbytes<0) {
			printf("Error sending msg: %d\n", nbytes);
		}
		else printf("%d bytes sent, spec=%d ms\n", nbytes, spec);

		//sleep 5 seconds between each service 
		Sleep(5000);
	}
	
	//close socket
	printf("Closing socket\n");
	error = closesocket(s);
	if (error!=0) {
		printf("Error closing socket=%d\n",error);
		getchar();
		return -1;
	}

	//close TCB session
	printf("Closing TCB session\n");
	error=close_tcb_session();
	if (error!=0) {
		printf("Error closing session=%d\n",error);
		getchar();
		return -1;
	}

	//wait for user to press a key
	printf("Press a key to close window...\n");
	getchar();
	return 0;
}